import { Debug, IDebug } from "./base";
import { JalhydObject } from "./jalhyd_object";
import { ParamDefinition } from "./param/param-definition";
import { IParamDefinitionIterator, ParamsEquation } from "./param/params-equation";

/**
 * type de calculette
 */
export enum CalculatorType {
    ConduiteDistributrice,
    LechaptCalmon,
    SectionParametree,
    RegimeUniforme,
    CourbeRemous,
    PabDimensions,      // passe à bassin rectangulaire
    PabPuissance,       // passe à bassin : puissance dissipée
    Structure,          // ouvrages hydrauliques simples
    ParallelStructure,  // ouvrages hydrauliques en parallèle
    Dever,              // Outil Cassiopée Dever
    Cloisons            // Outil Cassiopée PAB Cloisons
}

/**
 * type de noeud de calcul (sous type de calculette)
 */
export enum ComputeNodeType {
    None,
    // types de sections
    SectionTrapeze, SectionRectangle, SectionCercle, SectionPuissance,
    // types d'ouvrages hydrauliques
    StructureRectangle
}

/**
 * noeud de calcul
 */
// tslint:disable-next-line:max-classes-per-file
export abstract class ComputeNode extends JalhydObject implements IDebug {
    protected _prms: ParamsEquation;

    private _debug: Debug;

    constructor(prms: ParamsEquation, dbg: boolean = false) {
        super();
        this._debug = new Debug(dbg);
        this._prms = prms;
        this.debug("PARAMS", prms);
        if (!this._prms.calculabilityDefined) {
            this._prms.resetParametersCalculability();
        }
        this.setParametersCalculability();
        this._prms.DefineCalculability();
    }

    public get parameters(): ParamsEquation {
        return this._prms;
    }

    public getParameter(name: string): ParamDefinition {
        for (const p of this.parameterIterator) {
            if (p.symbol === name) {
                return p;
            }
        }
        return undefined;
    }

    public getFirstAnalyticalParameter(): ParamDefinition {
        for (const p of this.parameterIterator) {
            if (p.isAnalytical()) {
                return p;
            }
        }
        return undefined;
    }

    public get parameterIterator(): IParamDefinitionIterator {
        return this._prms.iterator;
    }

    // interface IDebug
    public debug(...args: any[]) {
        this._debug.debug(...args);
    }

    public get DBG(): boolean {
        return this._debug.DBG;
    }

    protected abstract setParametersCalculability(): void;

}
