import { Nub } from "../nub";
import { NamedIterableValues, NumberIterator } from "../param/param-value-iterator";
import { ParamValues } from "../param/param-values";
import { Result } from "../util/result";

/*
 * gestion de la valeur d'un paramètre par référence à une autre valeur (paramètre, résultat, résultat complémentaire)
 *
 * la référence (cf. INubReference.defineReference) (cf. INubReference.defineReference)
 * est de la forme <n | ouvrage[n] | N1>[.[N2]]
 *     n : indice de de l'ouvrage dans le cas des ouvrages parallèles
 *     N1 : un nom de paramètre/résultat (dans le cas d'un résultat, il est suivi d'un point)
 *     N2 : nom de résultat complémentaire (optionnel)
 *     ex :
 *       Q, Z1 (paramètres)
 *       J. (résultat)
 *       .Yf (résultat complémentaire du résultat courant)
 *       Q.Yf (résultat complémentaire du résultat nommé "Q")
 *       0.Q : paramètre Q du 1er ouvrage (ouvrages parallèles)
 *       ouvrage[1].Q_Mode : résultat complémentaire du 2ème ouvrage (ouvrages parallèles)
 */

/**
 * Nub dont certaines valeurs sont référençables pour réutilisation
 * (d'une calculette sur une autre par ex)
 */
export interface IReferencedNub {

    /**
     * getter des valeurs
     * @param desc : description sous forme symbolique
     */
    getReferencedParamValues(desc: string): ParamValues;

    /**
     * getter du résultat
     * @param desc : description sous forme symbolique
     */
    getReferencedResult(desc?: string): Result;

    /**
     * getter du résultat complémentaire
     * @param desc : description sous forme symbolique
     */
    getReferencedExtraResult(desc: string): any;

    /**
     * itérateur sur les valeurs
     */
    getReferencedValuesIterator(desc: string): NumberIterator;

    /**
     * objet (paramètre/résultat/résultat complémentaire) référencé
     */
    getReferencedObject(desc: string): NamedIterableValues;
}

/**
 * référence vers un Nub contenant une valeur vers laquelle on crée un lien
 */
export interface INubReference {

    /**
     * description symbolique de la référence
     */
    readonly referenceDefinition: string;

    /**
     * true si la référence a été définie
     */
    readonly isReferenceDefined: boolean;

    /**
     * instance de ParamValues référencée
     */
    readonly referencedParamValues: ParamValues;

    /**
     * instance de Result référencée
     */
    readonly referencedResult: Result;

    /**
     * instance de résultat complémentaire référencée
     */
    readonly referencedExtraResult: any;

    /**
     * itérateur sur les valeurs référencées
     */
    readonly referencedValuesIterator: NumberIterator;

    /**
     * objet (paramètre/résultat/résultat complémentaire) référencé
     */
    readonly referencedObject: NamedIterableValues;

    /**
     * définition de la valeur référencée dans le Nub
     * @param target Nub contenant la valeur qu'on va référencer
     * @param desc : description de la valeur pointée sous forme symbolique. Exemples : Q, <n° d'ouvrage>.Z1
     */
    defineReference(target: IReferencedNub, desc: string): void;

    /**
     * supprime la référence
     */
    undefineReference(): void;
}

/**
 * implémentation par défaut de INubReference
 */
export class NubReference implements INubReference {
    private _referencedNub: IReferencedNub;

    private _refDefinition: string;

    public defineReference(target: IReferencedNub, desc: string) {
        this._referencedNub = target;
        this._refDefinition = desc;
    }

    /**
     * supprime la référence
     */
    public undefineReference() {
        this._referencedNub = undefined;
        this._refDefinition = undefined;
    }

    /**
     * description symbolique de la référence
     */
    public get referenceDefinition(): string {
        return this._refDefinition;
    }

    public get isReferenceDefined(): boolean {
        return this._referencedNub != undefined && this._refDefinition != undefined;
    }

    /**
     * Nub référencé
     */
    public get referencedNub(): IReferencedNub {
        return this._referencedNub;
    }

    /**
     * instance de ParamValues référencée
     */
    public get referencedParamValues(): ParamValues {
        if (!this.isReferenceDefined) {
            return undefined;
        }
        const rpv = this._referencedNub.getReferencedParamValues(this._refDefinition);
        rpv._nubRef = this;
        return rpv;
    }

    /**
     * instance de Result référencée
     */
    public get referencedResult(): Result {
        if (!this.isReferenceDefined) {
            return undefined;
        }
        return this._referencedNub.getReferencedResult(this._refDefinition);
    }

    /**
     * instance de résultat complémentaire référencée
     */
    public get referencedExtraResult(): any {
        if (!this.isReferenceDefined) {
            return undefined;
        }
        return this._referencedNub.getReferencedExtraResult(this._refDefinition);
    }

    /**
     * itérateur sur les valeurs référencées
     */
    public get referencedValuesIterator(): NumberIterator {
        return this._referencedNub.getReferencedValuesIterator(this._refDefinition);
    }

    /**
     * objet (paramètre/résultat/résultat complémentaire) référencé
     */
    public get referencedObject(): NamedIterableValues {
        return this._referencedNub.getReferencedObject(this._refDefinition);
    }
}
